import pandas as pd

import provenance.serializers as s


def test_default_object_serializers():
    assert s.object_serializer('foo') == 'joblib'
    assert s.object_serializer((1, 2, 3)) == 'joblib'
    assert s.object_serializer({'foo': 42}) == 'joblib'

    df = pd.DataFrame([{'foo': 42}, {'foo': 55}])
    assert s.object_serializer(df) == 'pd_df_parquet'

    series = df.foo
    assert s.object_serializer(series) == 'pd_series_parquet'



# flake8: noqa
from ._config import from_config, load_config, load_yaml_config
from ._dependencies import dependencies
from ._version import get_versions
from .core import archive_file, ensure_proxies, promote, provenance, provenance_set
from .hashing import hash, value_repr
from .repos import (
    capture_set,
    create_set,
    current_config,
    get_check_mutations,
    get_default_repo,
    get_read_only,
    get_repo_by_name,
    get_use_cache,
    is_proxy,
    lazy_dict,
    lazy_proxy_dict,
    load_artifact,
    load_proxy,
    load_set_by_id,
    load_set_by_labels,
    load_set_by_name,
    set_check_mutations,
    set_default_repo,
    set_read_only,
    set_run_info_fn,
    set_use_cache,
    using_repo,
)
from .serializers import register_serializer

__version__ = get_versions()['version']
del get_versions



import graphviz
from frozendict import frozendict as fd

from ..repos import is_proxy


def elide(obj, length=30):
    table = str.maketrans({'{': r'\{', '}': r'\}', '<': r'\<', '>': r'\>'})
    s = str(obj).translate(table)
    return (s[:length] + '..') if len(s) > length else s


def artifact_id(artifact, length=7):
    return artifact.id[0:7]


def artifact_record(artifact, elide_len=30):
    return '|'.join(['<f0>' + artifact_id(artifact), '<f1>' + elide(artifact.value, elide_len)])


def param_node_id(child_artifact, name, val):
    if is_proxy(val):
        artifact = val.artifact
        return artifact.id
    # hmmm... we could share the inputs to other functions if we wanted to remove the child_artifact.id...
    return '|'.join([child_artifact.id, name])


def node(name, label=None, **attrs):
    attrs['type'] = 'node'
    attrs['name'] = name
    attrs['label'] = label
    return fd(attrs)


def edge(tail_name, head_name, **attrs):
    attrs['type'] = 'edge'
    attrs['tail_name'] = tail_name
    attrs['head_name'] = head_name
    return fd(attrs)


def dicts_to_digraph(dicts):
    g = graphviz.Digraph()
    for d in dicts:
        d = dict(d)
        t = d['type']
        del d['type']
        if t == 'node':
            g.node(**d)
        elif t == 'edge':
            g.edge(**d)
    return g


class DigraphDicts:

    def __init__(self):
        self.set = set()

    def node(self, name, label=None, **attrs):
        self.set.add(node(name, label, **attrs))
        return self

    def edge(self, tail_name, head_name, **attrs):
        self.set.add(edge(tail_name, head_name, **attrs))
        return self

    def to_dot(self):
        return dicts_to_digraph(self.set)

    def _repr_svg_(self):
        return self.to_dot()._repr_svg_()


def _viz_artifact(artifact, g):
    function_id = 'fn_' + artifact.id
    fn_qalified_name = '.'.join([artifact.fn_module, artifact.fn_name])
    fn_name = artifact.fn_name
    fn_params = '{fn}({params})'.format(
        fn=fn_qalified_name, params=','.join(artifact.inputs['kargs'].keys())
    )

    g.node(function_id, fn_name, shape='circle', tooltip=fn_params)
    g.edge(function_id, artifact.id)
    g.node(
        artifact.id,
        label=artifact_record(artifact, elide_len=15),
        shape='record',
        tooltip=elide(artifact.value, 50),
        color='red',
    )

    # ignore varargs for now...
    for name, val in artifact.inputs['kargs'].items():
        arg_node_id = param_node_id(artifact, name, val)
        if is_proxy(val):
            _viz_artifact(val.artifact, g)
            g.edge(val.artifact.id, function_id, label=name)
        else:
            g.node(arg_node_id, label=elide(val), shape='box')
            g.edge(arg_node_id, function_id, label=name)


def lineage_dot(artifact):
    """Walks the lineage of an artifact returning a DigraphDicts object
    that can be turned into a graphviz.Digraph and is automatically rendered
    as SVG in an IPython notebook.
    """
    g = DigraphDicts()
    if is_proxy(artifact):
        artifact = artifact.artifact
    _viz_artifact(artifact, g)
    return g



import os

import paramiko

from .. import blobstores as bs


def _ssh_client(ssh_config):
    client = paramiko.SSHClient()
    client.load_host_keys(os.path.expanduser('~/.ssh/known_hosts'))
    # There still seems to be problems with some types keys.
    # See https://github.com/paramiko/paramiko/issues/243
    # So you might try uncommenting if you are using an ecdsa-sha2-nistp256
    # client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    client.connect(**ssh_config)
    return client


class SFTPStore(bs.RemoteStore):

    def __init__(
        self,
        cachedir,
        basepath,
        ssh_config=None,
        ssh_client=None,
        sftp_client=None,
        read=True,
        write=True,
        read_through_write=True,
        delete=False,
        on_duplicate_key='skip',
        cleanup_cachedir=False,
        always_check_remote=False,
    ):
        """
        Parameters
        ----------
        always_check_remote : bool
           When True the SFTP server will be checked with every __contains__ call. Otherwise it will
        short-circuit if the blob is found in the cachedir. For performance reasons this
        should always be set to False. The only reason why you would want to use this
        is if you are using a SFTPStore and a DiskStore in a ChainedStore together for
        some reason. Since the SFTPStore basically doubles as a DiskStore with it's cachedir
        chaining the two doesn't really make sense though.
        """
        super(SFTPStore, self).__init__(
            always_check_remote=always_check_remote,
            cachedir=cachedir,
            basepath=basepath,
            cleanup_cachedir=cleanup_cachedir,
            read=read,
            write=write,
            read_through_write=read_through_write,
            delete=delete,
            on_duplicate_key=on_duplicate_key,
        )

        self.ssh_client = None
        if ssh_config is not None:
            self.ssh_client = _ssh_client(ssh_config)
        if self.ssh_client is not None:
            sftp_client = paramiko.SFTPClient.from_transport(self.ssh_client._transport)
        if sftp_client is not None:
            self.sftp_client = sftp_client
        else:
            # This is to allow testing the importing/subpackage aspect without
            # having to actually test the class by mocking an ssh connection.
            if cachedir is None and basepath is None:
                return
            raise ValueError(
                'You must specify a SFTP client by passing in one of: sftp_client, ssh_config, ssh_client'
            )

    def _exists(self, path):
        try:
            self.sftp_client.stat(path)
            return True
        except FileNotFoundError:
            return False

    def _delete_remote(self, path):
        self.sftp_client.remove(path)

    def _upload_file(self, filename, path):
        self.sftp_client.put(filename, path)

    def _download_file(self, remote_path, dest_filename):
        self.sftp_client.get(remote_path, dest_filename)



from __future__ import with_statement

from alembic import context
from sqlalchemy import engine_from_config, pool

from provenance import models

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

# add your model's MetaData object here
# for 'autogenerate' support
# from myapp import mymodel
# target_metadata = mymodel.Base.metadata
target_metadata = models.Base.metadata

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.


def run_migrations_offline():
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    url = config.get_main_option('sqlalchemy.url')
    context.configure(url=url, target_metadata=target_metadata, literal_binds=True)

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online():
    connectable = config.attributes.get('connection', None)

    if connectable is None:
        # only create Engine if we don't have a Connection
        # from the outside
        connectable = engine_from_config(
            config.get_section(config.config_ini_section),
            prefix='sqlalchemy.',
            poolclass=pool.NullPool,
        )

    # when connectable is already a Connection object, calling
    # connect() gives us a *branched connection*.

    with connectable.connect() as connection:
        context.configure(connection=connection, target_metadata=target_metadata)

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



import datetime
import os
import shutil
import time
from collections import namedtuple
from copy import copy

import toolz as t
from boltons import funcutils as bfu

from . import artifact_hasher as ah, repos as repos, serializers as s, utils
from ._dependencies import dependencies
from .hashing import file_hash, hash


class ImpureFunctionError(Exception):
    pass


class MutatedArtifactValueError(Exception):
    pass


def get_metadata(f):
    if hasattr(f, '_provenance_metadata'):
        return f._provenance_metadata
    if hasattr(f, 'func'):
        return get_metadata(f.func)
    else:
        return {}


artifact_properties = [
    'id',
    'value_id',
    'inputs',
    'fn_module',
    'fn_name',
    'value',
    'name',
    'version',
    'composite',
    'value_id_duration',
    'serializer',
    'load_kwargs',
    'dump_kwargs',
    'compute_duration',
    'hash_duration',
    'computed_at',
    'custom_fields',
    'input_artifact_ids',
    'run_info',
]

ArtifactRecord = namedtuple('ArtifactRecord', artifact_properties)


def fn_info(f):
    info = utils.fn_info(f)
    metadata = get_metadata(f)
    name = metadata['name'] or '.'.join([info['module'], info['name']])
    info['identifiers'] = {
        'name': name,
        'version': metadata['version'],
        'input_hash_fn': metadata['input_hash_fn'],
    }
    info['input_process_fn'] = metadata['input_process_fn']
    info['composite'] = metadata['returns_composite']
    info['archive_file'] = metadata['archive_file']
    info['custom_fields'] = metadata['custom_fields']
    info['preserve_file_ext'] = metadata['preserve_file_ext']
    info['use_cache'] = metadata['use_cache']
    info['read_only'] = metadata['read_only']
    if info['composite']:
        if info['archive_file']:
            raise NotImplementedError("Using 'composite' and 'archive_file' is not supported.")
        info['serializer'] = metadata['serializer'] or {}
        info['load_kwargs'] = metadata['load_kwargs'] or {}
        info['dump_kwargs'] = metadata['dump_kwargs'] or {}
        valid_serializer = isinstance(info['serializer'], dict)
        for serializer in info['serializer'].values():
            valid_serializer = valid_serializer and serializer in s.serializers
            if not valid_serializer:
                break
    elif info['archive_file']:
        serializer = metadata['serializer'] or 'file'
        if serializer != 'file':
            raise ValueError("With 'archive_file' set True the only valid 'serializer' is 'file'")
        if metadata.get('dump_kwargs') is not None:
            raise ValueError("With 'archive_file' set True you may not specify any dump_kwargs.")
        if metadata.get('load_kwargs') is not None:
            raise ValueError("With 'archive_file' set True you may not specify any load_kwargs.")
        info['serializer'] = 'file'
        info['load_kwargs'] = metadata['load_kwargs'] or {}
        info['dump_kwargs'] = metadata['dump_kwargs'] or {
            'delete_original': metadata['delete_original_file']
        }
        info['delete_original_file'] = metadata['delete_original_file']
        valid_serializer = True
    else:
        info['serializer'] = metadata.get('serializer', 'auto') or 'auto'
        info['load_kwargs'] = metadata.get('load_kwargs', None)
        info['dump_kwargs'] = metadata.get('dump_kwargs', None)
        valid_serializer = (info['serializer'] == 'auto' or info['serializer'] in s.serializers)

    if not valid_serializer:
        msg = 'Invalid serializer option "{}" for artifact "{}", available serialziers: {} '.format(
            info['serializer'], info['identifiers']['name'], tuple(s.serializers.keys())
        )
        raise ValueError(msg)

    return info


def hash_inputs(inputs, check_mutations=False, func_info=None):
    kargs = {}
    varargs = []
    all_artifacts = {}
    if func_info is None:
        func_info = {}

    for k, v in inputs['kargs'].items():
        h, artifacts = hash(v, hasher=ah.artifact_hasher())
        kargs[k] = h
        for a in artifacts:
            comp = all_artifacts.get(a.id, (a, []))
            comp[1].append(k)
            all_artifacts[a.id] = comp

    for i, v in enumerate(inputs['varargs']):
        h, artifacts = hash(v, hasher=ah.artifact_hasher())
        varargs.append(h)
        for a in artifacts:
            comp = all_artifacts.get(a.id, (a, []))
            comp[1].append('varargs[{}]'.format(i))
            all_artifacts[a.id] = comp

    if check_mutations:
        for comp in all_artifacts.values():
            a, arg_names = comp
            if a.value_id != hash(a.value):
                msg = 'Artifact {}, of type {} was mutated before being passed to {}.{} as arguments ({})'
                msg = msg.format(
                    a.id,
                    type(a.value),
                    func_info.get('module'),
                    func_info.get('name'),
                    ','.join(arg_names),
                )
                raise MutatedArtifactValueError(msg)

    input_hashes = {'kargs': kargs, 'varargs': tuple(varargs)}
    return (input_hashes, frozenset(all_artifacts.keys()))


def create_id(input_hashes, input_hash_fn, name, version):
    return t.thread_first(
        input_hashes, input_hash_fn, (t.merge, {
            'name': name,
            'version': version
        }), hash
    )


@t.curry
def composite_artifact(
    repo,
    _run_info,
    inputs,
    input_hashes,
    input_artifact_ids,
    input_hash_fn,
    artifact_info,
    compute_duration,
    computed_at,
    use_cache,
    read_only,
    key,
    value,
):
    start_hash_time = time.time()
    info = copy(artifact_info)
    info['composite'] = False
    info['name'] = '{}_{}'.format(info['name'], key)
    info['serializer'] = info['serializer'].get(key, 'auto')
    info['load_kwargs'] = info['load_kwargs'].get(key, None)
    info['dump_kwargs'] = info['dump_kwargs'].get(key, None)

    if info['serializer'] == 'auto':
        info['serializer'] = s.object_serializer(value)

    id = create_id(input_hashes, input_hash_fn, info['name'], info['version'])
    hash_duration = time.time() - start_hash_time
    value_id = hash(value)
    value_id_duration = time.time() - start_hash_time

    if not use_cache:
        id = hash(id + value_id)
    try:
        artifact = repo.get_by_id(id)
    except KeyError:
        record = ArtifactRecord(
            id=id,
            value_id=value_id,
            value=value,
            input_artifact_ids=input_artifact_ids,
            value_id_duration=value_id_duration,
            compute_duration=compute_duration,
            hash_duration=hash_duration,
            computed_at=computed_at,
            inputs=inputs,
            run_info=_run_info,
            **info
        )
        if read_only:
            artifact = repos._artifact_from_record(repo, record)
        else:
            artifact = repo.put(record)

    return artifact


def _base_fn(f):
    if utils.is_curry_func(f):
        return utils.inner_function(f)
    else:
        return f


_EXT_MAPPINGS = {'mpeg': 'mpg', 'jpeg': 'jpg'}


def _extract_extension(filename):
    ext = os.path.splitext(filename)[1]
    if len(ext) > 0:
        ext = ext.lower().strip()[1:]
        return '.' + _EXT_MAPPINGS.get(ext, ext)
    else:
        return ext


def _archive_file_hash(filename, preserve_file_ext):
    if hasattr(filename, '__fspath__'):
        filename = filename.__fspath__()
    else:
        filename = str(filename)
    if not os.path.exists(filename):
        raise FileNotFoundError(
            "Unable to archive file, {}, because it doesn't exist!".format(filename)
        )
    # TODO: figure out best place to put the hash_name config and use in both cases
    value_id = file_hash(filename)
    if preserve_file_ext:
        extension = _extract_extension(filename)
        value_id += extension
    return value_id


def run_info():
    return repos.Config.current().run_info()


@t.curry
def provenance_wrapper(repo, f):
    base_fn = _base_fn(f)
    extract_args = utils.args_extractor(base_fn, merge_defaults=True)
    func_info = fn_info(f)
    input_process_fn = func_info['input_process_fn']

    artifact_info = {
        'name': func_info['identifiers']['name'],
        'version': func_info['identifiers']['version'],
        'fn_name': func_info['name'],
        'fn_module': func_info['module'],
        'custom_fields': func_info['custom_fields'],
        'serializer': func_info['serializer'],
        'load_kwargs': func_info['load_kwargs'],
        'dump_kwargs': func_info['dump_kwargs'],
        'composite': func_info['composite'],
    }

    @bfu.wraps(f)
    def _provenance_wrapper(*args, **kargs):
        artifact_info_ = copy(artifact_info)
        r = repo
        if repo is None:
            r = repos.get_default_repo()
        elif isinstance(repo, str):
            r = repos.get_repo_by_name(repo)

        _run_info = run_info()
        archive_file = func_info['archive_file']

        if func_info['use_cache'] is None:
            use_cache = repos.get_use_cache()
        else:
            use_cache = func_info['use_cache']
        if func_info['read_only'] is None:
            read_only = repos.get_read_only()
        else:
            read_only = func_info['read_only']

        start_hash_time = time.time()
        varargs, argsd = extract_args(args, kargs)
        raw_inputs = {
            'varargs': varargs + func_info['varargs'],
            'kargs': t.merge(argsd, func_info['kargs']),
        }
        inputs = input_process_fn(raw_inputs)

        value_id = None
        filename = None
        archive_file_helper = (archive_file and '_archive_file_filename' in raw_inputs['kargs'])
        if archive_file_helper:
            filename = raw_inputs['kargs']['_archive_file_filename']
            value_id = _archive_file_hash(filename, func_info['preserve_file_ext'])
            inputs['filehash'] = value_id

        input_hashes, input_artifact_ids = hash_inputs(
            inputs, repos.get_check_mutations(), func_info
        )

        id = create_id(input_hashes, **func_info['identifiers'])
        hash_duration = time.time() - start_hash_time

        if use_cache:
            try:
                artifact = r.get_by_id(id)
            except KeyError:
                artifact = None
            except AttributeError as e:
                msg = 'The default repo is not set. '
                msg += 'You may want to add  the `default_repo` key to your call to `provenance.load_config.` '
                msg += "e.g., provenance.load_config({'default_repo': <default repo name>, ...})"
                raise AttributeError(msg).with_traceback(e.__traceback__)
        else:
            artifact = None

        if artifact is None:
            start_compute_time = time.time()
            computed_at = datetime.datetime.utcnow()
            value = f(*varargs, **argsd)
            compute_duration = time.time() - start_compute_time

            post_input_hashes, _ = hash_inputs(inputs)
            if id != create_id(post_input_hashes, **func_info['identifiers']):
                modified_inputs = []
                kargs = input_hashes['kargs']
                varargs = input_hashes['varargs']
                for name, _hash in post_input_hashes['kargs'].items():
                    if _hash != kargs[name]:
                        modified_inputs.append(name)
                for i, _hash in enumerate(post_input_hashes['varargs']):
                    if _hash != varargs[i]:
                        modified_inputs.append('varargs[{}]'.format(i))
                msg = 'The {}.{} function modified arguments: ({})'.format(
                    func_info['module'], func_info['name'], ','.join(modified_inputs)
                )
                raise ImpureFunctionError(msg)

            if artifact_info_['composite']:
                input_hash_fn = func_info['identifiers']['input_hash_fn']
                ca = composite_artifact(
                    r,
                    _run_info,
                    inputs,
                    input_hashes,
                    input_artifact_ids,
                    input_hash_fn,
                    artifact_info,
                    compute_duration,
                    computed_at,
                    use_cache,
                    read_only,
                )
                value = {k: ca(k, v) for k, v in value.items()}
                artifact_info_['serializer'] = 'auto'
                artifact_info_['load_kwargs'] = None
                artifact_info_['dump_kwargs'] = None

            if artifact_info_['serializer'] == 'auto':
                artifact_info_['serializer'] = s.object_serializer(value)

            start_value_id_time = time.time()
            if archive_file:
                if not archive_file_helper:
                    filename = value
                    value_id = _archive_file_hash(filename, func_info['preserve_file_ext'])
                value = ArchivedFile(value_id, filename, in_repo=False)
            else:
                value_id = hash(value)
            value_id_duration = time.time() - start_value_id_time

            if not use_cache:
                id = hash(id + value_id)
                try:
                    artifact = r.get_by_id(id)
                except KeyError:
                    artifact = None

            if artifact is None:
                record = ArtifactRecord(
                    id=id,
                    value_id=value_id,
                    value=value,
                    input_artifact_ids=input_artifact_ids,
                    value_id_duration=value_id_duration,
                    compute_duration=compute_duration,
                    hash_duration=hash_duration,
                    computed_at=computed_at,
                    run_info=_run_info,
                    inputs=inputs,
                    **artifact_info_
                )
                if read_only:
                    artifact = repos._artifact_from_record(r, record)
                else:
                    artifact = r.put(record)

            if archive_file:
                # mark the file as in the repo (yucky, I know)
                artifact.value.in_repo = True

        elif archive_file_helper and func_info.get('delete_original_file', False):
            # if we hit an artifact with archive_file we may still need to clean up the
            # referenced file. This is normally taken care of when the file is 'serialzied'
            # (see file_dump), but in the case of an artifact hit this would never happen.
            # One potential downside of this approach is that this local file will be
            # deleted and if the artifact value (i.e. the existing file) is not local
            # yet it will download the file that we just deleted. Another approach would
            # be to do a a put_overwrite which would potentially upload files multiple times.
            # So for now, the cleanest way is to accept the potential re-downloading of data.
            os.remove(filename)

        return artifact.proxy()

    if utils.is_curry_func(f):
        fb = bfu.FunctionBuilder.from_func(utils.inner_function(f))
        for arg in f.args + tuple(f.keywords.keys()):
            fb.remove_arg(arg)
        param_info = utils.param_info(f)
        args = []
        defaults = []
        for arg, value in param_info.items():
            args.append(arg)
            if value != utils.UNSPECIFIED_ARG:
                defaults.append(value)
        arg_inv = ['{}={}'.format(arg, arg) for arg in args]
        fb.body = 'return _provenance_wrapper(%s)' % ', '.join(arg_inv)
        fb.args = args
        fb.defaults = tuple(defaults)
        execdict = {'_provenance_wrapper': _provenance_wrapper}
        ret = fb.get_func(execdict, with_dict=True)
        return ret
    return _provenance_wrapper


def remove_inputs_fn(to_remove):

    def remove_inputs(inputs):
        kargs = t.keyfilter(lambda k: k not in to_remove, inputs['kargs'])
        return {'kargs': kargs, 'varargs': inputs['varargs']}

    return remove_inputs


def ensure_proxies(*parameters):
    """Decorator that ensures that the provided parameters are always arguments of type ArtifactProxy.

    When no parameters are passed then all arguments will be checked.

    This is useful to use on functions where you want to make sure artifacts
    are being passed in so lineage can be tracked.
    """

    def decorator(func):
        base_fn = _base_fn(func)
        extract_args = utils.args_extractor(base_fn, merge_defaults=True)

        @bfu.wraps(func)
        def check_args(*args, **kargs):
            _varargs, argsd = extract_args(args, kargs)
            not_valid = None
            if len(parameters) == 0:
                not_valid = [p for p, a in argsd.items() if not repos.is_proxy(a)]
            else:
                not_valid = [p for p in parameters if not repos.is_proxy(argsd[p])]
            if len(not_valid) > 0:
                msg = 'Arguments must be `ArtifactProxy`s but were not: [{}]'.format(
                    ', '.join(not_valid)
                )
                raise ValueError(msg)

            return func(*args, **kargs)

        return check_args

    return decorator


def provenance(
    version=0,
    repo=None,
    name=None,
    merge_defaults=None,
    ignore=None,
    input_hash_fn=None,
    remove=None,
    input_process_fn=None,
    archive_file=False,
    delete_original_file=False,
    preserve_file_ext=False,
    returns_composite=False,
    custom_fields=None,
    serializer=None,
    load_kwargs=None,
    dump_kwargs=None,
    use_cache=None,
    read_only=None,
    tags=None,
    _provenance_wrapper=provenance_wrapper,
):
    """
    Decorates a function so that all inputs and outputs are cached. Wraps the return
    value in a proxy that has an artifact attached to it allowing for the provenance
    to be tracked.


    Parameters
    ----------
    version : int
        Version of the code that is computing the value. You should increment this
        number when anything that has changed to make a previous version of an artifact
        outdated. This could be the function itself changing, other functions or libraries
        that it calls has changed, or an underlying data source that is being queried has
        updated data.

    repo : Repository or str
        Which repo this artifact should be saved in. The default repo is used when
        none is provided and this is the recommended approach. When you pass in a string
        it should be the name of a repo in the currently registered config.

    name : str
       The name of the artifact of the function being wrapped. If not provided it
       defaults to the function name (without the module).

    returns_composite : bool
       When set to True the function should return a dictionary. Each value of the
       returned dict will be serialized as an independent artifact. When the composite
       artifact is returned as a cached value it will be a dict-like object that will
       lazily pull back the artifacts as requested. You should use this when you need
       multiple artifacts created atomically but you do not want to fetch all the them
       simultaneously. That way you can lazily load only the artifacts you need.

    serializer : str
       The name of the serializer you want to use for this artifact. The built-in
       ones are 'joblib' (the default) and 'cloudpickle'. 'joblib' is optimized for
       numpy while 'cloudpickle' can serialize functions and other objects the standard
       python (and joblib) pickler cannot. You can also register your own serializer
       via the provenance.register_serializer function.

    dump_kwargs : dict
       A dict of kwargs to be passed to the serializer when dumping artifacts
       associated with this function. This is rarely used.

    load_kwargs : dict
       A dict of kwargs to be passed to the serializer when loading artifacts
       associated with this function. This is rarely used.

    ignore : list, tuple, or set
       A list of parameters that should be ignored when computing the input hash.
       This way you can mark certain parameters as invariant to the computed result.
       An example of this would be a parameter indicating how many cores should be
       used to compute a result. If the result is invariant the number of cores you
       would want to ignore it so the value isn't recomputed when a different number
       of cores is used.

    remove : list, tuple, or set
       A list of parameters that should be removed prior to hashing and saving
       of the inputs. The distinction between this and the ignore parameter is
       that with the ignore the parameters the ignored parameters are still recorded.
       The motivation to not-record, i.e. remove, certain parameters usually driven
       by performance or storage considerations.

    input_hash_fn : function
        A function that takes a dict of all on the argument's hashes with the
        structure of {'kargs': {'param_a': '1234hash'}, 'varargs': ('deadbeef',..)}.
        It should return a dict of the same shape but is able to change this dict
        as needed.  The main use case for this function is overshadowed by the
        ignore parameter and so this parameter is hardly ever used.

    input_process_fn : function
        A function that pre-processes the function's inputs before they are hashed
        or saved. The function takes a dict of all on the functions arguments with the
        structure of {'kargs': {'param_a': 42}, 'varargs': (100,..)}.
        It should return a dict of the same shape but is able to change this dict
        as needed.  The main use case for this function is overshadowed by the
        remove parameter and the value_repr function.

    merge_defaults : bool or list of parameters to be merged
        When True then the wrapper introspects the argspec of the function being
        decorated to see what keyword arguments have default dictionary values. When
        a list of strings the list is taken to be the list of parameters you want to
        merge on.
        When a decorated function is called then the dictionary passed in as an
        argument is merged with the default dictionary. That way people only need
        to specify the keys they are overriding and don't have to specify all the
        default values in the default dictionary.

    use_cache : bool or None (default None)
        use_cache False turns off the caching effects of the provenance decorator,
        while still tracking the provenance of artifacts. This should only be used during
        quick local iterations of a function to avoid having to bump the version with
        each change. When set to None (the default) it defers to the global provenance
        use_cache setting.

    read_only: bool or None (default None)
        read_only True will prevent any artifacts from being persisted to the repo.
        This should be used when you want to load existing artifacts from provenance
        but you do not want to add artifacts if the one you're looking for does not
        exist. This is useful when consuming artifacts created elsewhere, or when
        you are doing quick iterations (as with use_cache False) but you still want
        to use the cache for existing artifacts. When set to None (the default) it
        defers to the global provenance read_only setting.

    custom_fields : dict
        A dict with types that serialize to json. These are saved for searching in
        the repository.

    tags : list, tuple or set
        Will be added to custom_fields as the value for the 'tags' key.

    archive_file : bool, defaults False
       When True then the return value of the wrapped function will be assumed to
       be a str or pathlike that represents a file that should be archived into
       the blobstore. This is a good option to use when the computation of a function
       can't easily be returned as an in-memory pickle-able python value.

    delete_original_file : bool, defaults False
       To be used in conjunction with archive_file=True, when delete_original_file
       is True then the returned file will be deleted after it has been archived.

    preserve_file_ext : bool, default False
       To be used in conjunction with archive_file=True, when preserve_file_ext is
       True then id of the artifact archived will be the hash of the file contents
       plus the file extension of the original file. The motivation of setting this to
       True would be if you wanted to be able to look at the contents of a blobstore
       on disk and being able to preview the contents of an artifact with your
       regular OS tools (e.g. viewing images or videos).

    Returns
    -------
    ArtifactProxy
        Returns the value of the decorated function as a proxy. The proxy
        will act exactly like the original object/value but will have an
        artifact method that returns the Artifact associated with the value.
        This wrapped value should be used with all other functions that are wrapped
        with the provenance decorator as it will help track the provenance and also
        reduce redundant storage of a given value.
    """
    if ignore and input_hash_fn:
        raise ValueError('You cannot provide both ignore and input_hash_fn')

    if ignore:
        ignore = frozenset(ignore)
        input_hash_fn = remove_inputs_fn(ignore)

    if not input_hash_fn:
        input_hash_fn = lambda inputs: inputs

    if remove and input_process_fn:
        raise ValueError('You cannot provide both remove and input_process_fn')

    if remove:
        remove = frozenset(remove)
        input_process_fn = remove_inputs_fn(remove)

    if not input_process_fn:
        input_process_fn = lambda inputs: inputs

    def wrapped(f):
        _custom_fields = custom_fields or {}
        if tags:
            _custom_fields['tags'] = tags
        f._provenance_metadata = {
            'version': version,
            'name': name,
            'archive_file': archive_file,
            'delete_original_file': delete_original_file,
            'input_hash_fn': input_hash_fn,
            'input_process_fn': input_process_fn,
            'archive_file': archive_file,
            'delete_original_file': delete_original_file,
            'preserve_file_ext': preserve_file_ext,
            'returns_composite': returns_composite,
            'archive_file': archive_file,
            'custom_fields': _custom_fields,
            'serializer': serializer,
            'load_kwargs': load_kwargs,
            'dump_kwargs': dump_kwargs,
            'use_cache': use_cache,
            'read_only': read_only,
        }
        f.__merge_defaults__ = merge_defaults
        return _provenance_wrapper(repo, f)

    return wrapped


class ArchivedFile(object):

    def __init__(self, id, original_filename=None, in_repo=True):
        self.blob_id = id
        self.original_filename = original_filename
        self.in_repo = in_repo

    def abspath(self):
        repo = repos.get_default_repo()
        path = repo._filename(self.blob_id)
        return os.path.abspath(path)

    def __fspath__(self):
        return self.abspath() if self.in_repo else self.original_filename

    def __str__(self):
        return self.__fspath__()

    def __repr__(self):
        if self.original_filename:
            return '<ArchivedFile {}, {} >'.format(self.blob_id, self.original_filename)
        else:
            return '<ArchivedFile {} >'.format(self.blob_id)


def file_dump(archived_file, dest_filename, delete_original=False):
    op = shutil.move if delete_original else shutil.copy
    op(archived_file.original_filename, dest_filename)


def file_load(id):
    return ArchivedFile(id, in_repo=True)


s.register_serializer('file', file_dump, file_load)


def archive_file(
    filename, name=None, delete_original=False, custom_fields=None, preserve_ext=False
):
    """(beta) Copies or moves the provided filename into the Artifact Repository so it can
    be used as an ``ArtifactProxy`` to inputs of other functions.


    Parameters
    ----------
    archive_file : bool, defaults False
       When True then the return value of the wrapped function will be assumed to
       be a str or pathlike that represents a file that should be archived into
       the blobstore. This is a good option to use when the computation of a function
       can't easily be returned as an in-memory pickle-able python value.

    delete_original : bool, defaults False
       When delete_original_file True the file will be deleted after it has been archived.

    preserve_file_ext : bool, default False
       When True then id of the artifact archived will be the hash of the file contents
       plus the file extension of the original file. The motivation of setting this to
       True would be if you wanted to be able to look at the contents of a blobstore
       on disk and being able to preview the contents of an artifact with your
       regular OS tools (e.g. viewing images or videos).
    """

    # we want artifacts created by archive_file to be invariant to the
    # filename (see remove) but not the custom_fields.
    # filename is still passed in so the hash of the file on disk can be
    # computed as part of the id of the artifact.
    @provenance(
        archive_file=True,
        name=name or 'archive_file',
        preserve_file_ext=preserve_ext,
        delete_original_file=delete_original,
        remove=['_archive_file_filename'],
        custom_fields=custom_fields,
    )
    def _archive_file(_archive_file_filename, custom_fields):
        return filename

    return _archive_file(filename, custom_fields)


def provenance_set(set_labels=None, initial_set=None, set_labels_fn=None):
    if set_labels and set_labels_fn:
        raise ValueError('You cannot provide both set_labels and set_labels_fn.')

    def make_wrapper(f):
        if set_labels_fn:
            base_fn = _base_fn(f)
            extract_args = utils.args_extractor(base_fn, merge_defaults=True)
            func_info = utils.fn_info(f)

        @bfu.wraps(f)
        def wrapper(*fargs, **fkargs):
            if set_labels_fn:
                varargs, argsd = extract_args(fargs, fkargs)
                varargs += func_info['varargs']
                argsd.update(func_info['kargs'])
                labels = set_labels_fn(*varargs, **argsd)
            else:
                labels = set_labels

            with repos.capture_set(labels=labels, initial_set=initial_set) as result:
                f(*fargs, **fkargs)
            return result[0]

        return wrapper

    return make_wrapper


def promote(artifact_or_id, to_repo, from_repo=None):
    from_repo = from_repo if from_repo else repos.get_default_repo()
    artifact = repos.coerce_to_artifact(artifact_or_id, repo=from_repo)
    for a in dependencies(artifact):
        if a not in to_repo:
            to_repo.put(a)



import copy
import logging

import toolz as t

import provenance.blobstores as bs
import provenance.repos as r

logger = logging.getLogger(__name__)


@t.curry
def full_config(configs, base_config):
    if 'type' in base_config:
        return base_config
    prototype = full_config(configs, configs[base_config['prototype']])
    return t.thread_first(prototype, (t.merge, base_config), (t.dissoc, 'prototype'))


def merge_prototypes(config):
    return t.valmap(full_config(config), config)


@t.curry
def atomic_item_from_config(config, type_dict, item_plural, name=None):
    stype = config['type']
    if stype not in type_dict:
        raise Exception(
            '{} may only be created of types: {}, you had {}'.format(
                item_plural, tuple(type_dict.keys()), stype
            )
        )
    cls = type_dict[stype]
    kargs = t.dissoc(config, 'type')
    return cls(**kargs)


BLOBSTORE_TYPES = {
    'disk': bs.DiskStore,
    's3': bs.S3Store,
    'memory': bs.MemoryStore,
    'chained': bs.ChainedStore,
}

try:
    import provenance.sftp as sftp

    BLOBSTORE_TYPES['sftp'] = sftp.SFTPStore

except ImportError as e:

    class SFTPStore:
        _err = e

        def __init__(self, *args, **kargs):
            raise (self._err)

    BLOBSTORE_TYPES['sftp'] = SFTPStore

try:
    import provenance.google_storage as gs

    BLOBSTORE_TYPES['gs'] = gs.GSStore

except ImportError as e:

    class GSStore:
        _err = e

        def __init__(self, *args, **kargs):
            raise (self._err)

    BLOBSTORE_TYPES['gs'] = GSStore

blobstore_from_config = atomic_item_from_config(type_dict=BLOBSTORE_TYPES, item_plural='Blobstores')

REPO_TYPES = {
    'postgres': r.PostgresRepo,
    'memory': r.MemoryRepo,
    'chained': r.ChainedRepo,
}

repo_from_config = atomic_item_from_config(type_dict=REPO_TYPES, item_plural='Artifact Repos')


def items_from_config(config, atomic_from_config, items_name, item_type, silence_warnings):
    config = merge_prototypes(copy.deepcopy(config))

    atomic_stores = {}
    for k, c in config.items():
        try:
            if c['type'] != 'chained':
                store = atomic_from_config(c, name=k)
                if store:
                    atomic_stores[k] = store
        except Exception:
            if not silence_warnings:
                logger.warning(
                    'Error creating %s %s from config - Skipping',
                    item_type,
                    k,
                    exc_info=True,
                )

    def create_chained(name, config):
        # resolve the stores
        chained = {n for n in config[items_name] if n in atomic_stores}
        if len(chained) != len(config[items_name]):
            missing_configs = set(config[items_name]) - chained
            if not silence_warnings:
                logger.warning(
                    'Skipping chained %s %s due to missing %s: %s',
                    item_type,
                    name,
                    items_name,
                    missing_configs,
                )
            return None

        config[items_name] = [atomic_stores[n] for n in config[items_name]]
        return atomic_from_config(config, name=name)

    chained_stores = {}
    for k, c in config.items():
        try:
            if c['type'] == 'chained':
                store = create_chained(k, c)
                if store:
                    chained_stores[k] = store
        except Exception:
            if not silence_warnings:
                logger.warning(
                    'Error creating %s %s from config - Skipping',
                    item_type,
                    k,
                    exc_info=True,
                )

    return t.merge(chained_stores, atomic_stores)


def blobstores_from_config(config, silence_warnings=False):
    return items_from_config(config, blobstore_from_config, 'stores', 'blobstore', silence_warnings)


def repos_from_config(config, blobstores, silence_warnings=False):

    def from_config(atomic_config, name):
        if 'store' in atomic_config:
            if not atomic_config['store'] in blobstores:
                if not silence_warnings:
                    logger.warning(
                        'Skipping %s repo due to missing store: %s',
                        name,
                        atomic_config['store'],
                    )
                return None

            atomic_config['store'] = blobstores[atomic_config['store']]
        return repo_from_config(atomic_config)

    return items_from_config(config, from_config, 'repos', 'repo', silence_warnings)


def from_config(config):
    silence_warnings = config.get('silence_warnings', False)
    blobstores = blobstores_from_config(config['blobstores'], silence_warnings)
    repos = repos_from_config(config['artifact_repos'], blobstores, silence_warnings)
    return {'blobstores': blobstores, 'repos': repos}


def load_config(config):
    objs = from_config(config)
    pconfig = r.Config(
        objs['blobstores'],
        objs['repos'],
        default_repo=config['default_repo'],
        run_info_fn=config.get('run_info_fn', None),
        use_cache=config.get('use_cache', True),
        read_only=config.get('read_only', False),
        check_mutations=config.get('check_mutations', False),
    )
    r.Config.set_current(pconfig)
    return pconfig


def load_yaml_config(filename):
    import yaml

    with open(filename, 'r') as f:
        return load_config(yaml.load(f))



from boltons import funcutils as bfu
from google.cloud import storage as gs
from memoized_property import memoized_property

from . import blobstores as bs

# TODO: catch and retry w/new client on
# BrokenPipeError: [Errno 32] Broken pipe
# ConnectionResetError: [Errno 54] Connection reset by peer
# more?


def retry(f, max_attempts=2):

    @bfu.wraps(f)
    def with_retry(store, *args, **kargs):
        actual_attempts = 0
        while True:
            try:
                return f(store, *args, **kargs)
            except (BrokenPipeError, ConnectionError) as e:
                actual_attempts += 1
                if actual_attempts >= max_attempts:
                    raise e
                else:
                    store._setup_client()

    return with_retry


class GSStore(bs.RemoteStore):

    def __init__(
        self,
        cachedir,
        bucket,
        basepath='',
        project=None,
        read=True,
        write=True,
        read_through_write=True,
        delete=False,
        on_duplicate_key='skip',
        cleanup_cachedir=False,
        always_check_remote=False,
    ):
        """
        Parameters
        ----------
        always_check_remote : bool
           When True GS (Google Storage) will be checked with every __contains__ call. Otherwise it will
        short-circuit if the blob is found in the cachedir. For performance reasons this
        should always be set to False. The only reason why you would want to use this
        is if you are using a GSStore and a DiskStore in a ChainedStore together for
        some reason. Since the GSStore basically doubles as a DiskStore with it's cachedir
        chaining the two doesn't really make sense though.
        """
        super(GSStore, self).__init__(
            always_check_remote=always_check_remote,
            cachedir=cachedir,
            basepath=basepath,
            cleanup_cachedir=cleanup_cachedir,
            read=read,
            write=write,
            read_through_write=read_through_write,
            delete=delete,
            on_duplicate_key=on_duplicate_key,
        )

        self.bucket_name = bucket
        self.project = project

    def _setup_client(self):
        del self._client
        del self._bucket
        # force re-memoization
        assert self.bucket is not None

    @memoized_property
    def client(self):
        return gs.Client(project=self.project)

    @memoized_property
    def bucket(self):
        return self.client.get_bucket(self.bucket_name)

    @retry
    def _exists(self, path):
        blobs = list(self.bucket.list_blobs(prefix=path))
        return len(blobs) == 1

    @retry
    def _delete_remote(self, path):
        self.blob(path).delete()

    def _blob(self, path):
        return self._bucket.blob(path)

    @retry
    def _upload_file(self, filename, path):
        self._blob(path).upload_from_filename(filename)

    @retry
    def _download_file(self, remote_path, dest_filename):
        self._blob(remote_path).download_to_filename(dest_filename)



import sys

from . import hashing as h, repos as r


def _save(obj, artifacts):
    if isinstance(obj, r.Artifact):
        artifacts[obj.id] = obj
    if r.is_proxy(obj):
        artifacts[obj.artifact.id] = obj.artifact


class ArtifactHasher(h.Hasher):

    def __init__(self, artifacts=None, hash_name='md5'):
        if artifacts is None:
            artifacts = {}

        self.artifacts = artifacts
        h.Hasher.__init__(self, hash_name=hash_name)

    def save(self, obj):
        _save(obj, self.artifacts)
        h.Hasher.save(self, obj)

    def hash(self, obj):
        return (h.Hasher.hash(self, obj), self.artifacts.values())


class NumpyArtifactHasher(h.NumpyHasher):

    def __init__(self, artifacts=None, hash_name='md5', coerce_mmap=True):
        if artifacts is None:
            artifacts = {}

        self.artifacts = artifacts
        h.NumpyHasher.__init__(self, hash_name=hash_name, coerce_mmap=coerce_mmap)

    def save(self, obj):
        _save(obj, self.artifacts)
        h.NumpyHasher.save(self, obj)

    def hash(self, obj):
        return (h.NumpyHasher.hash(self, obj), self.artifacts.values())


def artifact_hasher(*args, **kwargs):
    if 'numpy' in sys.modules:
        return NumpyArtifactHasher(*args, **kwargs)
    else:
        return ArtifactHasher(*args, **kwargs)



from collections import namedtuple
from functools import singledispatch

import cloudpickle
import joblib
import toolz as t

from .hashing import hash


def cloudpickle_dump(obj, filename, **kwargs):
    with open(filename, 'wb') as f:
        return cloudpickle.dump(obj, f, **kwargs)


def cloudpickle_load(filename, **kwargs):
    with open(filename, 'rb') as f:
        return cloudpickle.load(f, **kwargs)


Serializer = namedtuple(
    'Serializer',
    'name, dump, load, content_type, content_encoding, content_disposition',
)


def joblib_dump(obj, filename, compress=2, **kwargs):
    joblib.dump(obj, filename, compress=compress, **kwargs)


serializers = {}


@singledispatch
def object_serializer(obj):
    """
    Takes an object and returns the appropirate serializer name, dump, and load arguments.

    Parameters
    ----------
    obj : any python object or primitive

    Returns
    -------
    tuple of serializer name (str), dump args (dictionary), load args (dictionary)
    """
    return DEFAULT_VALUE_SERIALIZER.name


def register_serializer(
    name,
    dump,
    load,
    content_type=None,
    content_encoding=None,
    content_disposition=None,
    classes=None,
):
    serializers[name] = Serializer(
        name, dump, load, content_type, content_encoding, content_disposition
    )
    if classes is None:
        return
    for cls in classes:
        object_serializer.register(cls, lambda _: name)


register_serializer('joblib', joblib_dump, joblib.load)
register_serializer('cloudpickle', cloudpickle_dump, cloudpickle_load)


def _pandas_and_parquet_present():
    try:
        import pandas    # noqa: F401
    except ImportError:
        return False
    try:
        import pyarrow    # noqa: F401
    except:
        try:
            import fastparquet    # noqa: F401
        except ImportError:
            return False
    return True


if _pandas_and_parquet_present():
    import pandas as pd

    def pd_df_parquet_dump(df, filename, **kwargs):
        return df.to_parquet(filename, **kwargs)

    def pd_df_parquet_load(filename, **kwargs):
        return pd.read_parquet(filename, **kwargs)

    register_serializer(
        'pd_df_parquet', pd_df_parquet_dump, pd_df_parquet_load, classes=[pd.DataFrame]
    )

    def pd_series_parquet_dump(series, filename, **kwargs):
        if series.name is None:
            # pyarrow requires the column names be strings
            series = pd.Series(series, name='_series')
        return pd.DataFrame(series).to_parquet(filename, **kwargs)

    def pd_series_parquet_load(filename, **kwargs):
        series = pd.read_parquet(filename, **kwargs).ix[:, 0]
        if series.name == '_series':
            series.name = None
        return series

    register_serializer(
        'pd_series_parquet',
        pd_series_parquet_dump,
        pd_series_parquet_load,
        classes=[pd.Series],
    )


def _pytorch_present():
    try:
        import torch    # noqa F401
    except:
        return False
    return True


if _pytorch_present():
    import torch

    def pytorch_model_dump(model, filename, **kwargs):
        return torch.save(model, filename)

    def pytorch_model_load(filename, **kwargs):
        return torch.load(filename)

    register_serializer(
        'pytorch_model',
        pytorch_model_dump,
        pytorch_model_load,
        classes=[torch.nn.Module],
    )


@t.memoize(key=lambda *args: hash(args))
def partial_serializer(serializer_name, dump_kwargs, load_kwargs):
    s = serializers[serializer_name]
    return Serializer(
        s.name,
        t.partial(s.dump, **dump_kwargs) if dump_kwargs else s.dump,
        t.partial(s.load, **load_kwargs) if load_kwargs else s.load,
        s.content_type,
        s.content_encoding,
        s.content_disposition,
    )


def serializer(artifact):
    return partial_serializer(artifact.serializer, artifact.dump_kwargs, artifact.load_kwargs)


DEFAULT_VALUE_SERIALIZER = serializers['joblib']
DEFAULT_INPUT_SERIALIZER = serializers['joblib']



#!/usr/bin/env python

from os.path import exists

from setuptools import setup

import versioneer

subpackages = {
    'sftp': ['paramiko'],
    'google_storage': ['google-cloud'],
    'vis': ['graphviz', 'frozendict'],
}

DESCRIPTION = 'Provenance and caching library for functions, built for creating lightweight machine learning pipelines.'

setup(
    name='provenance',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    packages=['provenance', 'provenance.sftp', 'provenance.vis'],
    install_requires=[open('requirements.txt').read().strip().split('\n')],
    extras_require=subpackages,
    include_package_data=True,
    description=DESCRIPTION,
    long_description=(open('README.rst').read() if exists('README.rst') else ''),
    author='Ben Mabey',
    author_email='ben@benmabey.com',
    url='http://github.com/bmabey/provenance',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
    ],
)



import contextlib
import os
import os.path
import shutil
import tempfile

from joblib.disk import mkdirp
from s3fs import S3FileSystem

from . import _commonstore as cs
from .serializers import DEFAULT_VALUE_SERIALIZER


class BaseBlobStore:

    def __init__(
        self,
        read=True,
        write=True,
        read_through_write=True,
        delete=False,
        on_duplicate_key='skip',
    ):
        self._read = read
        self._write = write
        self._read_through_write = read_through_write
        self._delete = delete
        self._on_duplicate_key = on_duplicate_key

        valid_on_duplicate_keys = {'skip', 'overwrite', 'check_collision', 'raise'}
        if self._on_duplicate_key not in valid_on_duplicate_keys:
            msg = 'on_duplicate_key must be one of {}'.format(valid_on_duplicate_keys)
            raise RuntimeError(msg)

    def __getitem__(self, id, *args, **kargs):
        return self.get(id, *args, **kargs)

    def put(self, id, value, serializer=DEFAULT_VALUE_SERIALIZER, read_through=False):
        method = getattr(self, '_put_' + self._on_duplicate_key)
        return method(id, value, serializer, read_through)

    def _put_raise(self, id, value, serializer, read_through):
        cs.ensure_put(self, id, read_through)
        self._put_overwrite(id, value, serializer, read_through)

    def _put_skip(self, id, value, serializer, read_through):
        if id not in self:
            self._put_overwrite(id, value, serializer, read_through)

    def _put_check_collision(self, id, value, serializer, read_through):
        cs.ensure_put(self, id, read_through, check_contains=False)
        if id not in self:
            self._put_overwrite(id, value, serializer, read_through)
        else:
            self._check_collision(id, value, serializer)

    # TODO: Right now our only thought is that this can be
    # checked by using an alternate hash, this will require
    # deserializing the old value and running the hash algorithm
    # with an alternate hash
    def _check_collision(self, id, value, serializer):
        raise NotImplementedError()


class MemoryStore(BaseBlobStore):

    def __init__(
        self,
        values=None,
        read=True,
        write=True,
        read_through_write=True,
        delete=True,
        on_duplicate_key='skip',
    ):
        super(MemoryStore, self).__init__(
            read=read,
            write=write,
            read_through_write=read_through_write,
            delete=delete,
            on_duplicate_key=on_duplicate_key,
        )
        if values is None:
            self.values = {}
        else:
            self.values = values

    def __contains__(self, id):
        cs.ensure_contains(self)
        return id in self.values

    def _put_overwrite(self, id, value, serializer, read_through):
        cs.ensure_put(self, id, read_through, check_contains=False)
        self.values[id] = value

    def get(self, id, serialzier=None, **_kargs):
        cs.ensure_read(self)
        cs.ensure_present(self, id)
        return self.values[id]

    def delete(self, id):
        cs.ensure_delete(self, id)
        del self.values[id]


@contextlib.contextmanager
def _temp_filename():
    try:
        temp = tempfile.NamedTemporaryFile('wb', delete=False)
        temp.close()
        yield temp.name
    finally:
        if os.path.isfile(temp.name):
            os.remove(temp.name)


@contextlib.contextmanager
def _atomic_write(filename):
    with _temp_filename() as temp:
        yield temp
        shutil.move(temp, filename)


def _abspath(path):
    return os.path.abspath(os.path.expanduser(path))


class DiskStore(BaseBlobStore):

    def __init__(
        self,
        cachedir,
        read=True,
        write=True,
        read_through_write=True,
        delete=False,
        on_duplicate_key='skip',
    ):
        super(DiskStore, self).__init__(
            read=read,
            write=write,
            read_through_write=read_through_write,
            delete=delete,
            on_duplicate_key=on_duplicate_key,
        )
        self.cachedir = _abspath(cachedir)
        mkdirp(self.cachedir)

    def _filename(self, id):
        return os.path.join(self.cachedir, id)

    def __contains__(self, id):
        cs.ensure_contains(self)
        return os.path.isfile(self._filename(id))

    def _put_overwrite(self, id, value, serializer, read_through):
        cs.ensure_put(self, id, read_through, check_contains=False)
        with _atomic_write(self._filename(id)) as temp:
            serializer.dump(value, temp)

    def get(self, id, serializer=DEFAULT_VALUE_SERIALIZER, **_kargs):
        cs.ensure_read(self)
        cs.ensure_present(self, id)
        return serializer.load(self._filename(id))

    def delete(self, id):
        cs.ensure_delete(self, id)
        os.remove(self._filename(id))


class RemoteStore(BaseBlobStore):

    def __init__(
        self,
        cachedir,
        basepath,
        read=True,
        write=True,
        read_through_write=True,
        delete=False,
        on_duplicate_key='skip',
        cleanup_cachedir=False,
        always_check_remote=False,
    ):
        """
        Parameters
        ----------
        always_check_remote : bool
           When True the remote store will be checked with every __contains__ call. Otherwise it will
        short-circuit if the blob is found in the cachedir. For performance reasons this
        should always be set to False. The only reason why you would want to use this
        is if you are using a RemoteStore and a DiskStore in a ChainedStore together for
        some reason. Since the RemoteStore basically doubles as a DiskStore with it's cachedir
        chaining the two doesn't really make sense though.
        """
        super(RemoteStore, self).__init__(
            read=read,
            write=write,
            read_through_write=read_through_write,
            delete=delete,
            on_duplicate_key=on_duplicate_key,
        )

        self.always_check = always_check_remote

        self.cachedir = _abspath(cachedir)
        self.basepath = basepath
        self.cleanup_cachedir = cleanup_cachedir
        mkdirp(self.cachedir)

    def __del__(self):
        if self.cleanup_cachedir:
            shutil.rmtree(self.cachedir)

    def _filename(self, id):
        return os.path.join(self.cachedir, id)

    def _path(self, id):
        return os.path.join(self.basepath, id)

    def _exists(self, path):
        raise NotImplementedError()

    def _delete_remote(self, path):
        raise NotImplementedError()

    def _upload_file(self, filename, path):
        raise NotImplementedError()

    def _download_file(self, path, dest_filename):
        raise NotImplementedError()

    def __contains__(self, id):
        cs.ensure_contains(self)
        path = self._path(id)
        if self.always_check:
            return self._exists(path)
        else:
            return os.path.exists(self._filename(id)) or self._exists(path)

    def _put_overwrite(self, id, value, serializer, read_through):
        cs.ensure_put(self, id, read_through, check_contains=False)
        filename = self._filename(id)
        # not already saved by DiskStore?
        if not os.path.isfile(filename):
            with _atomic_write(filename) as temp:
                serializer.dump(value, temp)
        self._upload_file(filename, self._path(id))

    def get(self, id, serializer=DEFAULT_VALUE_SERIALIZER, **_kargs):
        cs.ensure_read(self)
        cs.ensure_present(self, id)
        filename = self._filename(id)
        if not os.path.exists(filename):
            with _atomic_write(filename) as temp:
                self._download_file(self._path(id), temp)
        return serializer.load(filename)

    def delete(self, id):
        cs.ensure_delete(self, id)
        filename = self._filename(id)
        if os.path.exists(filename):
            os.remove(filename)
        self._delete_remote(self._path(id))


class S3Store(RemoteStore):

    def __init__(
        self,
        cachedir,
        basepath,
        s3_config=None,
        s3fs=None,
        read=True,
        write=True,
        read_through_write=True,
        delete=False,
        on_duplicate_key='skip',
        cleanup_cachedir=False,
        always_check_remote=False,
    ):
        """
        Parameters
        ----------
        always_check_remote : bool
           When True S3 will be checked with every __contains__ call. Otherwise it will
        short-circuit if the blob is found in the cachedir. For performance reasons this
        should always be set to False. The only reason why you would want to use this
        is if you are using a S3Store and a DiskStore in a ChainedStore together for
        some reason. Since the S3Store basically doubles as a DiskStore with it's cachedir
        chaining the two doesn't really make sense though.
        """
        super(S3Store, self).__init__(
            always_check_remote=always_check_remote,
            cachedir=cachedir,
            basepath=basepath,
            cleanup_cachedir=cleanup_cachedir,
            read=read,
            write=write,
            read_through_write=read_through_write,
            delete=delete,
            on_duplicate_key=on_duplicate_key,
        )

        if s3fs:
            self.s3fs = s3fs
        elif s3_config is not None:
            self.s3fs = S3FileSystem(**s3_config)
        else:
            raise ValueError('You must provide either s3_config or s3fs for a S3Store')

    def _exists(self, path):
        return self.s3fs.exists(path)

    def _delete_remote(self, path):
        self.s3fs.rm(path)

    def _upload_file(self, filename, path):
        self.s3fs.put(filename, path)

    def _download_file(self, remote_path, dest_filename):
        self.s3fs.get(remote_path, dest_filename)


class ChainedStore(BaseBlobStore):

    def __init__(
        self,
        stores,
        read=True,
        write=True,
        read_through_write=True,
        delete=True,
        on_duplicate_key='skip',
    ):
        super(ChainedStore, self).__init__(
            read=read,
            write=write,
            read_through_write=read_through_write,
            delete=delete,
            on_duplicate_key=on_duplicate_key,
        )
        self.stores = stores

    def __contains__(self, id):
        return cs.chained_contains(self, id)

    def _filename(self, id):
        return cs.chained_filename(self, id)

    def _put_overwrite(self, id, value, serializer, read_through):
        return cs.chained_put(self, id, value, overwrite=True, serializer=serializer)

    def get(self, id, serializer=DEFAULT_VALUE_SERIALIZER, **kargs):

        def get(store, id):
            return store.get(id, serializer=serializer, **kargs)

        return cs.chained_get(self, get, id)

    def __getitem__(self, id, **kargs):
        return self.get(id, **kargs)

    def delete(self, id):
        return cs.chained_delete(self, id)
